#ifndef LT_AUTOFLOOD_HH_
#define LT_AUTOFLOOD_HH_

//##############################################################################
//##############################################################################

#include "ell/LandscapeTopology.hh"
#include "ell/Basin.hh"

namespace ell
{


	  /*! An instance of this class wrappes another LandscapeTopology.
	   * 
	   * For each added minimum it floods the corresponding gradient basin up
	   * to a given maximal energy relative to the minimum's energy.
	   *  
	   * The minimal saddle found (if any) is used to perform a gradient walk
	   * into the neighbored basin to identify the local minimum that is 
	   * reachable via this lowest barrier. If the neighbored minimum has LOWER
	   * energy (is smaller according to State::less operator) 
	   * than the current one, the detected saddle between the minima is added. 
	   * If the minimum is not already part of the topology it is added
	   * before.
	   * 
	   * NOTE: This is a recursive scheme if the identified neighbored minimum 
	   * to add is not already part of the topology. The recursion ends if
	   * (a) no saddle is found within the energy threshold or (b) the 
	   * neighbored minimum reachable via the lowest saddle is known or not  
	   * smaller according to the State::less operator. 
	   * 
	   * @author Martin Mann
	   */
	class LT_AutoFlood : public LandscapeTopology
	{
	protected:
		
		  //! the wrapped landscape topology
		LandscapeTopology* lt;
		
		  //! the maximal energy height to flood according to the minimum 
		  //! started from
		double maxFloodHeight;

		  //! the delta E to lower maxE in case too many states are in the queue
		  //! during flooding (see Basin_F class)
		const double floodDeltaE;
		  //! the maximal number of elements to store in the queue during 
		  //! flooding (see Basin_F class)
		const size_t floodMaxToStore;
		
		  //! whether or not the lt should be deleted when life cycle of this
		  //! object ends
		const bool deleteLT;

	public:
		
		/*! Construction
		 * @param lt the wrapped landscape topology
		 * @param floodMaxHeight the relative maximal energy height to flood 
		 *        above each minimum added
		 * @param floodDeltaE the delta E to lower maxE in case too many states 
		 *        are in the queue during flooding (see Basin_F class)
		 * @param floodMaxToStore the maximal number of elements to store in 
		 *        the queue during flooding (see Basin_F class)
		 * @param deleteLT whether or not the lt should be deleted when life 
		 *        cycle of this object ends
		 */ 
		LT_AutoFlood(	LandscapeTopology * lt
						, const double floodMaxHeight
						, const double floodDeltaE
						, const size_t floodMaxToStore
						, const bool deleteLT = false );
		
		virtual ~LT_AutoFlood();
		
		//! Sets a saddle point that connects two minima 
		//! @param i index of the first minimum to connect
		//! @param j index of the second minimum to connect
		//! @param s the State representating the saddle point
		//! @return whether or not the data structure has been changed
		virtual 
		bool 
		addSaddle(const size_t i, const size_t j, const State* const s);
		
		//! Sets a saddle point that connects two minima 
		//! @param m_i the first minimum to connect
		//! @param m_j the second minimum to connect
		//! @param s the State representating the saddle point
		//! @return whether or not the data structure has been changed
		virtual 
		bool 
		addSaddle(const State* const m_i, const State* const m_j, const State* const s);
		
		//! Adds an UNKNOWN minimum to the wrapped topology.
		//! 
		//! FURTHERMORE : the basin of the minimum is flooded up to a given 
		//! maximal energy and the lowest saddle below that is identified.
		//! If existing the neighbored minimum is calculated using a gradient
		//! walk and the detected saddle between the minima is added.
		//! If the neighbored minimum was not part of the topology a recursive
		//! call of addMin(neighMin) is done before addSaddle is called.
		//!
		//! @param m the minimum to add
		//! @return the index of the minimum in the landscape
		virtual 
		const size_t 
		addMin(const State* const m);
		
		//! Access to the minimum index.
		//! @param m the minimum of that the index should be accessed
		//! @return  the index of the minimum, or INVALID_INDEX if 'm' is no known 
		//!          minimum
		virtual 
		const size_t 
		getMinIndex(const State* const m) const;
		
		//! Access to the number of stored minima.
		//! @return the number of minima
		virtual 
		const size_t 
		getMinCount() const;
		
		//! Access to a minimum.
		//! @param i index of the minimum to access
		//! @return the State with the specified minimum index
		virtual 
		const State* const 
		getMin(size_t i) const;
		
		//! Access to the global minimum of the wrapped landscape with minimal 
		//! energy.
		//! @return the minimum with smallest energy (and smallest string 
		//!         representation for tiebreaking)
		virtual 
		const State* const 
		getMFEState() const;

		//! Calculates the relative 'distance' between landscape topologies.
		//! @param lt the landscape topology to calculate the distance to
		//! @return the distance
		virtual 
		double 
		getDistance(const LandscapeTopology* const lt) const;
		
		//! Sorts the wrapped minima ascending by their energy (mfe minimum 
		//! gets index 0) using the State::less operator.
		virtual 
		void 
		sort();

		//! Check if the minima of the wrapped landscape are sorted ascendingly 
		//! by the State::less operator.
		virtual 
		bool 
		isSorted() const;

		//! Clears the wrapped topology to be filled again afterwards. 
		virtual 
		void 
		clear();

		//! Writes the wrapped landscape topology to stream.
		//! @param out the output stream to write to.
		//! @param writeGraph whether or not the output should contain a graph
		//!        representation of the topology in graphviz DOT format
		//! @param writeE prints the energy for each state after its string
		//!        representation
		virtual 
		void
		write(	std::ostream & out, 
				const bool writeGraph = false,
				const bool writeE = false ) const;

		//! Fills wrapped the landscape topology from stream.
		//! @param in the input stream to read from.
		//! @param templateState the State template to create new State objects via
		//!        its 'fromString' function
		//! @return an error encoding and an error string :
		//!          0 = no error occured, read was successfull
		//!         -1 = landscape topology type not supported
		//!         -2 = state description differs from the given one
		//!         -3 = other read errors
		virtual 
		std::pair<int,std::string>
		read(	std::istream & in, 
				const State * const templateState  );

	};

} // namespace ell

  // inline method definitions
#include "LT_AutoFlood.icc"


//##############################################################################
//##############################################################################

#endif /*LT_AUTOFLOOD_HH_*/
