#ifndef LT_BARRIERTREE_HH_
#define LT_BARRIERTREE_HH_

// get information on available hashes or maps
#include <biu/HashMap.hh>

// include best available hash or map
#if HAVE_UNORDERED_MAP == 1
	#include <unordered_map>
#else
#if HAVE_TR1_UNORDERED_MAP == 1
	#include <tr1/unordered_map>
#else
#if HAVE_GNU_HASH_MAP == 1
	#include <ext/hash_map>
#else
	#include <map>
#endif
#endif
#endif

#include "ell/LT_MinimaSet.hh"
#include "ell/State.hh"

#include <vector>
#include <iostream>
#include <string>


//##############################################################################
//##############################################################################


namespace ell
{

	/*! This class represents a barrier tree of an energy landscape.
	 * 
	 * A barrier tree is a hierachical representation of the minimal energy 
	 * barriers between minima. Thus leaves represent minima and internal nodes
	 * minimal barriers between the children of these nodes.
	 * 
	 * This implementation stores the minimal barrier state within the internal
	 * node.
	 * 
	 * It is possible that an object of this class represents a FOREST of
	 * barrier trees, i.e. it exists at least one pair of minima without a 
	 * stored barrier between them. Thus not all minima are connected to all 
	 * other via a barrier.
	 *
	 * Furthermore it allows for the aggregation of minima. A minimum M1 is
	 * aggregated to another minimum M2 if
	 * - M2 is smaller M1 (according to the State::less order)
	 * - it exists a barrier B between M1-M2 such that 
	 *   E(B)-E(M1) <= maxMergeBarrier (user defined)
	 * 
	 * In case that holds, the leaf of M1 is merged with the leaf containing M2.
	 * Furthermore the barriers of both leaves are merged to maintain the 
	 * barrier information of both minima. But M1 is NOT removed from the 
	 * minima list and still accessible via its old index!
	 * 
	 * When enabling minima merging (maxMergeBarrier >= 0.0) one has to note
	 * that the structure of the barrier tree might get depending on the order
	 * of the barrier input IF the inserted saddles are only barriers without
	 * adjacency information. IF that is the case, the minimum M2 is not ensured
	 * to be the NEXT minimum reachable by the barrier given but ANY that is 
	 * reachable by that barrier. 
	 * 
	 */
	class LT_BarrierTree: public LT_MinimaSet {
	
	public:
		
		  //! the landscape topology id string used in the output header to
		  //! identify a minima set
		static const std::string LT_ID;
		
		
	protected:
		
		class StateList;
		class NodeList;
		
		  //! Represents a node in the barrier tree that can store a list of
		  //! States that are either barriers (internal nodes) or minima 
		  //! (leaves).  
		class Node {
		public:
			  //! the parent node in the tree or NULL in case of the root node
			Node* parent;
			  //! the list of child nodes or NULL in case of leave nodes
			NodeList* children;
			  //! the list of states represented by this node
			StateList* content;
			
			  //! construction
			Node(	Node* parent, 
					NodeList* children, 
					StateList* content,
					const bool isRoot = false );

			  //! destruction
			~Node();
		};
		
		class NodeList {
		public:
			  //! the currently accessed node of the list
			Node* current;
			  //! the successor of this node in the list or NULL if this is the
			  //! end of the node list
			NodeList* next;
			
			  //! construction
			NodeList(	Node* current, NodeList* next );
			
			  //! destruction
			~NodeList();
			
			  //! adds the node toAdd to this list
			static
			void
			addToList( NodeList*& list, Node* toAdd );
			
			static
			void
			remFromList( NodeList*& listHead, const Node* toRem );
			
		};
		
		class StateList {
		public:
			  //! the currently accessed State of the list
			State* current;
			  //! the successor of this State in the list or NULL if this is the
			  //! end of the State list
			StateList* next;
			
			  //! construction
			StateList(	State* current, StateList* next );
			
			  //! destruction
			~StateList();
			
			  //! adds the State toAdd to this list
			static
			void
			addToList( StateList*& list, State* toAdd );
		};
		
		
		class hash_StateP {
		public:
			size_t operator()(const State* s) const {
				return (size_t)s;
			}
		};
		class hash_NodeP {
		public:
			size_t operator()(const Node* n) const {
				return (size_t)n;
			}
		};

		// set typedef for best available hash or map
		#if HAVE_UNORDERED_MAP == 1
			typedef std::unordered_map< const State*, Node*, hash_StateP > Min2Node_MAP;
			typedef std::unordered_map< const Node*, const Node*, hash_NodeP > Node2Node_MAP;
			typedef std::unordered_map< const Node*, size_t, hash_NodeP > Node2SizeT_MAP;
			typedef std::unordered_map< const Node*, int, hash_NodeP > Node2Int_MAP;
			typedef std::unordered_map< size_t, const Node* > SizeT2Node_MAP;
		#else
		#if HAVE_TR1_UNORDERED_MAP == 1
			typedef std::tr1::unordered_map< const State*, Node*, hash_StateP > Min2Node_MAP;
			typedef std::tr1::unordered_map< const Node*, const Node*, hash_NodeP > Node2Node_MAP;
			typedef std::tr1::unordered_map< const Node*, size_t, hash_NodeP > Node2SizeT_MAP;
			typedef std::tr1::unordered_map< const Node*, int, hash_NodeP > Node2Int_MAP;
			typedef std::tr1::unordered_map< size_t, const Node* > SizeT2Node_MAP;
		#else
		#if HAVE_GNU_HASH_MAP == 1
			typedef __gnu_cxx::hash_map< const State*, Node*, hash_StateP > Min2Node_MAP;
			typedef __gnu_cxx::hash_map< const Node*, const Node*, hash_NodeP > Node2Node_MAP;
			typedef __gnu_cxx::hash_map< const Node*, size_t, hash_NodeP > Node2SizeT_MAP;
			typedef __gnu_cxx::hash_map< const Node*, int, hash_NodeP > Node2Int_MAP;
			typedef __gnu_cxx::hash_map< size_t, const Node* > SizeT2Node_MAP;
		#else
			typedef std::map< const State*, Node* > Min2Node_MAP;
			typedef std::map< const Node*, const Node* > Node2Node_MAP;
			typedef std::map< const Node*, size_t > Node2SizeT_MAP;
			typedef std::map< const Node*, int > Node2Int_MAP;
			typedef std::map< size_t, const Node* > SizeT2Node_MAP;
		#endif
		#endif
		#endif


	protected:
		
		  //! the vector of all known minima
		using LT_MinimaSet::vMinima;
		  //! index of the global minimum (with minimal free energy)
		using LT_MinimaSet::mfeIndex;
		  //! internal flag that stores if vMinima is sorted or not
		using LT_MinimaSet::sorted;
		  //! internal mapping of vMinima string representations to their index
		using LT_MinimaSet::minMap;
		
		  //! the energy threshold used to merge minima to equivalence classes
		double maxMergeBarrier;
		  //! the root of the barrier tree
		Node* root;
		  //! mapping from minima to the barrier tree nodes
		Min2Node_MAP min2node;
		  //! number of leaves in the barrier tree
		size_t numOfLeaves;
		
	public:
		
		//! Empty construction.
		//! @param maxMergeBarrier the maximal barrier height of a minimum to 
		//!        a lower one to be merged into that leave.
		//! @param hashMinima whether or not we have to hash on the minima 
		//!        string representations to allow for a faster lookup within 
		//!        the getMinIndex(..) member function
		LT_BarrierTree(  const double maxMergeBarrier = 0.0,
						const bool hashMinima = true );
		
		//! Copy construction.
		LT_BarrierTree(  const LT_BarrierTree& toCopy );
		
		//! Destruction.
		virtual ~LT_BarrierTree();

		
		///////////////// LANDSCAPE TOPOLOGY INTERFACE //////////////////
		
		//! Sets a saddle point that connects two minima 
		//! @param i index of the first minimum to connect
		//! @param j index of the second minimum to connect
		//! @param s the State representating the saddle point
		//! @return whether or not the data structure has been changed
		virtual 
		bool 
		addSaddle(const size_t i, const size_t j, const State* const s);
		
		//! Sets a saddle point that connects two minima 
		//! @param m_i the first minimum to connect
		//! @param m_j the second minimum to connect
		//! @param s the State representating the saddle point
		//! @return whether or not the data structure has been changed
		virtual 
		bool 
		addSaddle(const State* const m_i, const State* const m_j, const State* const s);
		
		//! Adds an UNKNOWN minimum to the topology.
		//! NOTE : this operation should NOT change the indices of already 
		//!        existing minima in the topology !
		//! FURTHER : the minimum is not checked if it is already known
		//! @param m the first minimum to add
		//! @return the index of the minimum in the landscape
		virtual 
		const size_t 
		addMin(const State* const m);
		
		//! Calculates the relative 'distance' between two barrier trees. 
		//! It is done on the minima and barriers. For each
		//! minimum m present in only one of the two LTs a value of 
		//! MISSING_MINIMUM_PENALTY*(e^(-energy(m)/BOLTZMANN_KT)) is summed. The 
		//! resulting sum is normalized by the number of (unique) minima in both
		//! topologies.
		//! For all O(n^2) barriers in the trees the RMSD is calculated and 
		//! summed to the minima penalty. For each barrier present in one of the
		//! trees only, the maximal barrier of both trees (root(mfe_tree)-mfe) 
		//! is used in the RMSD calculation for the missing barrier.
		//! Note: The used 'dummy' barrier height is only the highest barrier 
		//! if the barrier tree is no forest (only mfe-tree considered) ...
		//! @param lt the barrier tree to calculate the distance to
		//! @return the distance
		virtual 
		double 
		getDistance(const LandscapeTopology* const lt) const;
		
		//! Clears the data structure and removes all stored information to be 
		//! filled again afterwards. 
		virtual 
		void 
		clear();

		//! Writes the current landscape topology to stream.
		//! @param out the output stream to write to.
		//! @param writeGraph whether or not the output should contain a graph
		//!        representation of the topology in graphviz DOT format
		//! @param writeE prints the energy for each state after its string
		//!        representation
		virtual 
		void
		write(	std::ostream & out, 
				const bool writeGraph = false,
				const bool writeE = false ) const;

		//! Fills the landscape topology from stream.
		//! @param in the input stream to read from.
		//! @param templateState the State template to create new State objects
		//!        via its 'fromString' function
		//! @return an error encoding and an error string :
		//!          0 = no error occured, read was successfull
		//!         -1 = landscape topology type not supported
		//!         -2 = state description differs from the given one
		//!         -3 = other read errors (see error string)
		virtual 
		std::pair<int,std::string>
		read(	std::istream & in, 
				const State * const templateState );
		
		
		//! Checks whether or not the barrier tree represents one tree or a
		//! forest of at least two unconnected trees.
		//! @return true if this tree represents more than one unconnected
		//!         barrier tree; false otherwise
		bool
		isForest( void ) const;
		
		//! Access for the number of trees this barrier tree represents.
		//! @return if isForest() == true, the number of trees, otherwise 1
		size_t
		getNumOfTrees( void ) const;
		
		//! Access to the number of leaves the tree holds. The number 
		//! corresponds to the number of minima cluster / equivalence classes.
		//! @return the number of tree leaves 
		size_t
		getNumOfLeaves( void ) const;
		
		//! Assignment operator.
		//! @param toCopy the barrier tree this object should be a copy of
		//! @return access to the changed this object
		virtual
		LT_BarrierTree&
		operator = ( const LT_BarrierTree& toCopy );
		
		//! Assignment operator.
		//! @param toCopy the minima set this object should be a copy of
		//! @return access to the changed this object
		virtual
		LT_BarrierTree&
		operator = ( const LT_MinimaSet& toCopy );
		
		
		//! Access to the smallest barrier state between two minima.
		//! @param m1 the index of the first minimum
		//! @param m2 the index of the second minimum
		//! @return the smallest barrier state known between two minima or
		//!         NULL if minima belong to not connected trees of a forest
		virtual
		const State*
		getBarrier( const size_t m1, const size_t m2 ) const;
		
		
		//! Access to the smallest barrier state between two minima.
		//! NOTE : the two minima have to be already part of the topology !!!
		//! @param m1 the first minimum 
		//! @param m2 the second minimum
		//! @return the smallest barrier state known between two minima or
		//!         NULL if minima belong to not connected trees of a forest
		virtual
		const State*
		getBarrier( const State* const m1, const State* const m2 ) const;
		
		
	protected:
		
		//! Access to the highest (grand) parent node in the tree of a given
		//! node that represents an energy below or equal to the given maximum.
		//! @param child the child node to start search from
		//! @param maxEnergy the energy maximum that should not be exceeded
		//! @return the highest anchester of the node below the energy maximum
		Node*
		getHighestParentLEQ( Node* child, const double maxEnergy );
		
		//! Constant Access to the highest (grand) parent node in the tree of a 
		//! given node that represents an energy below or equal to the given 
		//! maximum.
		//! @param child the child node to start search from
		//! @param maxEnergy the energy maximum that should not be exceeded
		//! @return the highest anchester of the node below the energy maximum
		const Node*
		getHighestParentLEQ( const Node* child, const double maxEnergy ) const;
		
		//! Merges recursively two subtrees until the tree is valid again. It
		//! returns the root node of the resulting merged subtree that has to
		//! be checked if it is obsolete afterwards (checkIfObsolete).
		Node*
		mergeSubTrees( Node* t1, Node* t2 );
		
		//! Access to the lowest common anchester of two nodes in the tree.
		//! @param n1 the first node of the query
		//! @param n2 the second node of the query
		//! @return the LCA of the two nodes in the tree
		Node*
		getLCA( Node* n1, Node* n2 );
		
		//! Constant access to the lowest common anchester of two nodes in the 
		//! tree.
		//! @param n1 the first node of the query
		//! @param n2 the second node of the query
		//! @return the LCA of the two nodes in the tree
		const Node*
		getLCA( const Node* n1, const Node* n2 ) const;
		
		//! Checks if an internal node has only one child node and is therefore
		//! obsolete in the tree. If so, the node is removed from the tree.
		//! NOTE: The node itself is NOT DELETED, only unlinked!
		//! @param n the node to check
		//! @return the final root of this subtree
		Node*
		checkIfObsolete( Node* n );
		
	public:
		
		/*!
		 * Writes the barrier tree in postscript format to stream for a viewable
		 * tree representation.
		 * 
		 * NOTE: Currently NO FORESTS are supported and no output will be 
		 * generated in case this object represents a forest.
		 * 
		 * The original source code was derived from the 'barriers' program by
		 * the TBI group of Vienna University.
		 * 
		 * NOTE: The barrier tree will be sorted to enable the printing!!! This
		 * might cause additional runtime and a changed minima indexing!
		 * 
		 * @param out the stream to write to
		 * @param maxLeaves2print maximal number of leaves to print
		 */
		void
		write_PStree(  std::ostream& out, const size_t maxLeaves2print = UINT_MAX ) ; 
		
	protected:  ///////////////  write_PStree stuff ////////////////////////////
		
		/*!
		 * Temporary data structure for write_PStree
		 */
		class nodeT {
		public:
			nodeT() : height(0.0), saddle_height(0.0), father(INT_MAX), label(NULL) {}
		  float height;         /* height (energy, time, whatever) of this leaf    */
		  float saddle_height;  /* height of internal node that connects this leaf */
		  int father;           /* node with which it connects                     */
		  char *label;          /* label string, if NULL use index+1               */ 
		};

		/*!
		 * Temporary data structure for write_PStree
		 */
		class linkT {
		public:
			linkT() : x(0), next(NULL) {}
		  int x;
		  linkT *next;
		};

		/*!
		 * Prints a barrier tree encoded in nodeT structures to stream.
		 * 
		 * The original source code was derived from the 'barriers' program by
		 * the TBI group of Vienna University.
		 * 
		 * @param out the stream to write to
		 * @param nodes the list of leaf nodes to print
		 */
		static void PS_tree_plot(	std::ostream& out
									, const std::vector<nodeT> &nodes
									);
		
		//! Helper function class of PS_tree_plot to sort the nodes by saddle 
		//! height.
		class cmp_saddle {
		public:
			//! construction
			cmp_saddle(const std::vector<nodeT> *leaves_) : leaves(leaves_) {}
			//! Access to the currently printed nodes to enable the sorting of the
			//! data structure via cmp_saddle.
			const std::vector<nodeT> *leaves;
			//! comparison function
			//! @param A first position in leaves
			//! @param B second position in leaves
			//! @return true if saddle height of A is smaller than of B, or if
			//!         equal if height of A is smaller than B
			bool operator() (int A,int B);
		};

	};

} // namespace ell

//##############################################################################

  // include inline implementation
#include "ell/LT_BarrierTree.icc"

//##############################################################################

#endif /*LT_BARRIERTREE_HH_*/
