
#include "ell/MinimumMapping.hh"

#include <ell/Walk.hh>
#include <ell/StateCollector.hh>

#include <biu/assertbiu.hh>


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


namespace ell {

	MM_GradientWalk::MM_GradientWalk()
	{
	}

	MM_GradientWalk::~MM_GradientWalk()
	{
	}

	State*
	MM_GradientWalk::
	getLocalMin( const State * const toMap ) const {
		assertbiu( toMap != NULL, "no State to map given" );
		
		  // set up state collector that will contain the minimum as the last 
		  // added state
		SC_Counting sc;
		  // perform the gradient walk
		WalkGradient::walkGradient(
						toMap, 
						&sc
						);
		  // return a clone of the last added State = local minimum
		return sc.getLastAdded()->clone();
	}

	
	size_t
	MM_GradientWalk::
	getLocalMinIndex(	const State * const toMap
						, const LandscapeTopology& lt ) const
	{
		assertbiu( toMap != NULL, "no State to map given" );

		  // set up state collector that will contain the minimum as the last 
		  // added state
		SC_Counting sc;
		  // perform the gradient walk
		WalkGradient::walkGradient(
						toMap, 
						&sc
						);
		  // return the index in lt of the last added State = local minimum
		return lt.getMinIndex(sc.getLastAdded());
	}

} // namespace ell



//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////



namespace ell {

	MM_GradientWalkHashed::
	MM_GradientWalkHashed( const size_t maxHashSize )
	 :	gradWalk( maxHashSize )
	{
	}

	MM_GradientWalkHashed::~MM_GradientWalkHashed()
	{
	}

	State*
	MM_GradientWalkHashed::
	getLocalMin( const State * const toMap ) const {
		assertbiu( toMap != NULL, "no State to map given" );
		
		  // set up state collector that will contain the minimum as the last 
		  // added state
		SC_Counting sc;
		  // perform the gradient walk
		gradWalk.walk(	toMap, 
						&sc,
						NULL
						);
		  // return a clone of the last added State = local minimum
		return sc.getLastAdded()->clone();
	}

	
	size_t
	MM_GradientWalkHashed::
	getLocalMinIndex(	const State * const toMap
						, const LandscapeTopology& lt ) const
	{
		assertbiu( toMap != NULL, "no State to map given" );

		  // set up state collector that will contain the minimum as the last 
		  // added state
		SC_Counting sc;
		  // perform the gradient walk
		gradWalk.walk(
						toMap, 
						&sc,
						NULL
						);
		  // return the index in lt of the last added State = local minimum
		return lt.getMinIndex(sc.getLastAdded());
	}

	
	void 
	MM_GradientWalkHashed::
	setHashSize( const size_t s ) { 
		gradWalk.setHashSize(s); 
	}

	const size_t
	MM_GradientWalkHashed::
	getHashSize( void ) { 
		return gradWalk.getHashSize(); 
	}

	void
	MM_GradientWalkHashed::
	clearHash( void ) {
		gradWalk.clearHash();
	}

	

} // namespace ell



//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


namespace ell {

	MM_Hashed_LT::
	MM_Hashed_LT(	const MinimumMapping& wrappedMapping_
					, const LandscapeTopology& allMinima_
					, const size_t maxHashSize )
	 :	wrappedMapping( wrappedMapping_ )
	 	, lt(allMinima_)
	 	, seenStates( maxHashSize )
	{
	}

	MM_Hashed_LT::
	~MM_Hashed_LT()
	{
	}

	State*
	MM_Hashed_LT::
	getLocalMin( const State * const toMap ) const 
	{
		assertbiu(toMap != NULL ,"no state to map given");
		
		// search in mapping
		State2MinMap::const_iterator found = seenStates.find(toMap->toString());
		
		// check if already hashed
		if (found != seenStates.end()) {
			  // return a clone of the mapped minimum
			return lt.getMin(found->second)->clone();
		} 
		
		// else not already hashed ..
		
		// get mapped minimum
		size_t minIndex = wrappedMapping.getLocalMinIndex( toMap, lt );
		
		assertbiu( minIndex != ell::LandscapeTopology::INVALID_INDEX
					, "mapped minimum is not part of the LandscapeTopology");
		
		// store mapping in hash
		seenStates.insert( State2MinMapVal(	toMap->toString(), minIndex ) );
		
		// return a copy of the mapped minimum
		return lt.getMin( minIndex )->clone();
		
	}

	size_t
	MM_Hashed_LT::
	getLocalMinIndex(	const State * const toMap
						, const LandscapeTopology& lt2 ) const 
	{
		assertbiu(toMap != NULL ,"no state to map given");
		
		// search in mapping
		State2MinMap::const_iterator found = seenStates.find(toMap->toString());
		
		// check if already hashed
		if (found != seenStates.end()) {
			  // return the index of the mapped minimum
			return found->second;
		} 
		
		// else not already hashed ..
		
		// get index of the mapped minimum
		size_t minIdx = wrappedMapping.getLocalMinIndex( toMap, lt );
		assertbiu( minIdx != ell::LandscapeTopology::INVALID_INDEX
					, "mapped minimum is not part of the internal LandscapeTopology");
		
		// store mapping in hash
		seenStates.insert( State2MinMapVal(	toMap->toString(), minIdx ) );
		
		return minIdx;
	}


} // namespace ell



//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


