#include "ell/RejectionlessMCWalk.hh"
#include <biu/RandomNumberFactory.hh>
#include <ell/LandscapeTopology.hh>

#include <cmath>

using namespace ell;

RejectionlessMCWalk::
SC_TimedWrapper::SC_TimedWrapper(StateCollector* _stateCollector)
	: 	lastTime(0),
		stateCollector(_stateCollector)
{}

void
RejectionlessMCWalk::
SC_TimedWrapper::add(const State& s) {
	stateCollector->add(s);
}

void
RejectionlessMCWalk::
SC_TimedWrapper::add(const State& s, const double endTime) {
	stateCollector->add(s);
	lastTime = endTime;
}

size_t
RejectionlessMCWalk::
SC_TimedWrapper::size() const {
	return stateCollector->size();
}

const State* const
RejectionlessMCWalk::
SC_TimedWrapper::getLastAdded() const {
	return stateCollector->getLastAdded();
}

double
RejectionlessMCWalk::
SC_TimedWrapper::getLastTime() const {
	return lastTime;
}

RejectionlessMCWalk::RejectionlessMCWalk() {
}

RejectionlessMCWalk::RejectionlessMCWalk(const RejectionlessMCWalk& orig) {
}

RejectionlessMCWalk::~RejectionlessMCWalk() {
}

void
RejectionlessMCWalk::walk(	const State& start,
							StateCollector* scWalk,
							double maxTime,
							const double kT,
							TransitionMetric metric) {

	// setup
	double time = 0;
    State* currentState = start.clone();
    SC_Timed* timedCollector = dynamic_cast<SC_Timed*>(scWalk);
    if (timedCollector==NULL) {
    	timedCollector = new SC_TimedWrapper(scWalk);
    }

    while (time < maxTime) {
    	// sum up all neighbors
    	double currentEnergy = currentState->getEnergy();
    	double totalFlux = 0;	// sum of all transition rates
    	State::NeighborListPtr neighlist = currentState->getNeighborList();
		State::NeighborList::Iterator it1 = neighlist->begin();
		assertbiu(it1 != neighlist->end(), "error: this state has no neighbors");
		while (it1 != neighlist->end()) {
			if (metric == KAWASAKI) {
				totalFlux += getKawasaki(currentEnergy, it1->getEnergy(), kT);
			}
			else {
				totalFlux += getMetropolis(currentEnergy, it1->getEnergy(), kT);
			}
			++it1;
		}

		// choose new neighbor
		// get random number in [0,totalFlux]
		double randomNumber = totalFlux *
			(double) biu::RandomNumberFactory::getRN() /
			 (double) biu::RandomNumberFactory::getMaxRN();
		State::NeighborList::Iterator it2 = neighlist->begin();
		double sum = 0;
		while (it2 != neighlist->end()) {
			if (metric == KAWASAKI) {
				sum += getKawasaki(currentEnergy, it2->getEnergy(), kT);
			} else {
				sum += getMetropolis(currentEnergy, it2->getEnergy(), kT);
			}
			if (sum >= randomNumber) break;
			++it2;
		}
		assertbiu(it2 != neighlist->end(), "algorithm is not allowed to reach"
				"end of neighborlist");

		// compute progressed time
		// get random number in [0,1]
		randomNumber = (double) biu::RandomNumberFactory::getRN() /
						 (double) biu::RandomNumberFactory::getMaxRN();
		double timeIncrement = log(1./randomNumber) / totalFlux;
		time += timeIncrement;

		timedCollector->add(*currentState, time);

    	// resume walk with new neighbor
    	delete currentState;
    	currentState = (*it2).clone();
    }
    delete currentState;
}

double
ell::RejectionlessMCWalk::getMetropolis(double energyOrigin,
										double energySucc,
										double kT)
{
	if (energySucc <= energyOrigin) {
		return 1;
	}
	else {
		return std::exp(-(energySucc-energyOrigin)/kT);
	}
}

double
ell::RejectionlessMCWalk::getKawasaki(	double energyOrigin,
										double energySucc,
										double kT) {
	return std::exp( -(energySucc-energyOrigin)/(2*kT) );
}


