#ifndef STATEACCEPTOR_HH_
#define STATEACCEPTOR_HH_


#include "ell/State.hh"
#include "ell/StateCollector.hh"

namespace ell {
	/*!
	 * This abstract class is used to check if a given State is
	 * valid to be accepted as successor in a walk.
	 * 
	 * @author Daniel Maticzka
	 */	
	class StateAcceptor 
	{
	public:	
		
		virtual ~StateAcceptor(){};
		
		/*!
		 * Returns true, if State succ is a valid successor to
		 * the states colleted by StateCollector sc.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const = 0;
	};
	
	/*!
	 * This subclass of StateAcceptor accepts all states as successor in a walk.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_All : public StateAcceptor
	{
	public:
		
		virtual ~SA_All() {};
		
		/*!
		 * Always returns true.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return returns always true
		 */
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
	/*!
	 * This subclass of StateAcceptor accepts a given State as successor if
	 * it is smaller according to the less operator '<' of the last accepted
	 * State in the StateCollector.
	 * 
	 * @author Martin Mann
	 */	
	class SA_Le : public StateAcceptor
	{
	public:
		
		virtual ~SA_Le() {}
		
		/*!
		 * Tests if succ is smaller than the last accepted State of 
		 * StateCollector sc using succ.operator<(*sc->getLastAdded()).
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor 
		 */	
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
	/*!
	 * This subclass of StateAcceptor accepts a given State as successor if
	 * it has equal energy to the last accepted State in the StateCollector.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_E_Eq : public StateAcceptor
	{
	public:
		
		virtual ~SA_E_Eq() {}
		
		/*!
		 * Tests for equal energy of State succ and last accepted State in
		 * StateCollector sc.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */		
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
	/*!
	 * This subclass of StateAcceptor accepts a given State as successor if
	 * it has higher energy than the last accepted State in the StateCollector.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_E_Gr : public StateAcceptor
	{
	public:
		
		virtual ~SA_E_Gr() {}
		
		/*!
		 * Tests if energy of State succ is higher than energy of last accepted
		 * State of StateCollector sc.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */		
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
	/*!
	 * This subclass of StateAcceptor accepts a given State as successor if
	 * it has higher or equal energy than the last accepted State in the 
	 * StateCollector.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_E_GrEq : public StateAcceptor
	{
	public:
		
		virtual ~SA_E_GrEq () {};
		
		/*!
		 * Tests if energy of State succ is greater or equal to energy of last 
		 * accepted State of StateCollector sc.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */	
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
	/*!
	 * This subclass of StateAcceptor accepts a given State as successor if
	 * it has smaller energy than the last accepted State in the 
	 * StateCollector.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_E_Le : public StateAcceptor
	{
	public:
		
		virtual ~SA_E_Le() {}
		
		/*!
		 * Tests if energy of State succ is less than energy of last 
		 * accepted State of StateCollector sc.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */	
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
	/*!
	 * This subclass of StateAcceptor accepts a given State as successor if
	 * it has smaller or equal energy than the last accepted State in the 
	 * StateCollector.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_E_LeEq : public StateAcceptor
	{
	public:
		
		virtual ~SA_E_LeEq() {}
		
		/*!
		 * Tests if energy of State succ is less or equal to energy of last 
		 * accepted State of StateCollector sc.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */	
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
	
	/*!
	 * This subclass of StateAcceptor accepts a given State as successor if
	 * it has smaller or equal energy than the given maximal value.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_Max_E : public StateAcceptor
	{
	public:
		
		  //! construction
		  //! @param MAX_ENERGY the maximal energy allowed for accepted states
		SA_Max_E( const double MAX_ENERGY );
		
		virtual ~SA_Max_E() {}
		
		/*!
		 * Tests if energy of State succ is less or equal to MAX_ENERGY.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */	
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
		
	protected:
		
		  //! the maximal energy allowed for accepted states.
		const double MAX_ENERGY;
	};
	
	
	/*!
	 * This subclass of StateAcceptor
	 * implements the Metropolis criterion to determine if a given State is
	 * valid to be accepted as successor in a walk.
	 * The acceptance probability used is min( 1, e^(-deltaE/kT) ) utilizing
	 * the Boltzmann weight of energy difference deltaE between the new State 
	 * and the last accepted one in the state collector.
	 * 
	 * @author Daniel Maticzka
	 */	
	class SA_Metropolis : public StateAcceptor {
	protected:
		
		  //! the weighted temperature to calculate the Boltzmann weight of a
		  //! State.
		const double kT;

	public:
		
		  //! Construction
		  //! @param kT the weighted temperature to calculate the Boltzmann 
		  //! weight of a State.
		SA_Metropolis(const double kT);

		virtual ~SA_Metropolis() {}
		
		/*!
		 * Checks if State succ is valid successor to last accepted State
		 * from StateCollector sc, using the metropolis criterion.
		 * @param sc the StateCollector to test against
		 * @param succ the successor State to test
		 * @return true if valid successor
		 */
		virtual bool accept(
				const StateCollector* const sc, 
				const State& succ) const;
	};
	
} // namespace ell

#endif /*STATEACCEPTOR_HH_*/
