#ifndef STATEPAIRCOLLECTOR_HH_
#define STATEPAIRCOLLECTOR_HH_


//##############################################################################
//##############################################################################

#include "ell/State.hh"

#include <utility> // contains std::pair definition

namespace ell 
{
	/*!
	 * This abstract interface is used to keep track of visited  
	 * pairs of states. For instance, it is used during flooding of the 
	 * landscape to determine the contact planes between basins. 
	 * 
	 * @author Hannes Kochniss
	 * @author Martin Mann
	 */
	class StatePairCollector
	{
	public:	
		
		  //! a pair of States
		typedef std::pair<const State*,const State*> StatePair;
	
		virtual
		~StatePairCollector() {}
		
		
		  //! This function makes the collector manage the given State pair.
		  //! @param s the State pair to add 
		virtual void add(const StatePair& s) = 0;
		
		  //! This function makes the collector manage the given State pair.
		  //! @param s1 the first State of the pair to add 
		  //! @param s2 the second State of the pair to add 
		virtual void add(	const State* const s1
							, const State* const s2) = 0;
		
		  //! Returns number of added State pairs.
		  //! @return the number of added state pairs
		virtual size_t size() const = 0;
		
		  //! Returns last added State pair.
		  //! Note: The pointed States will be deleted when the collector is
		  //! destroyed. So a cloning is needed in case the pair will be needed
		  //! later.
		  //! @return last added State pair
		virtual const StatePair getLastAdded() const = 0;
		
	};
	
} // namespace ell

//##############################################################################
//##############################################################################


namespace ell
{
	/*!
	 * This class is used to keep track of the number of States pairs
	 * added to the collector
	 * 
	 * @author Hannes Kochniss
	 * @author Martin Mann
	 */
	class SPC_Counting : public StatePairCollector
	{
		
	protected:
	
		  //! the first State of the last added State pair 
		State* 	lastAdded_s1;
		  //! the second State of the last added State pair 
		State* 	lastAdded_s2;
		  //! the number of added States seen
		size_t 		statePairCount;
		
	public:	
	
		SPC_Counting();
		
		virtual
		~SPC_Counting();
	
		  //! Increments State pair counter and stores a copy this pair as the
		  //! last added.
		  //! @param s the State pair to add
		virtual void add(	const StatePair& s );
		
		  //! Increments State pair counter and stores this pair as the last
		  //! added.
		  //! @param s1 the first State of the pair to add 
		  //! @param s2 the second State of the pair to add 
		virtual void add(	const State* const s1
							, const State* const s2 );
		
		  //! Returns number of added State pairs.
		  //! @return the number of added state pairs
		virtual size_t size() const;
		 
		  //! Returns last added State pair.
		  //! Note: The pointed States will be deleted when the collector is
		  //! destroyed. So a cloning is needed in case the pair will be needed
		  //! later.
		  //! @return last added State pair
		virtual const StatePair getLastAdded() const;
		
	protected:
		
		  //! delete the last added state pair
		void clearLastAdded();
		
	};
	
	  // inline defintion 
	inline void SPC_Counting::clearLastAdded() {
		if (lastAdded_s1 != NULL)
			delete lastAdded_s1;
		if (lastAdded_s2 != NULL)
			delete lastAdded_s2;
	}

	
} // namespace ell


//##############################################################################
//##############################################################################

#include <list>

namespace ell
{
	/*!
	 * This class stores an accessible list of all State pairs added to the 
	 * collector.
	 * 
	 * @author Hannes Kochniss
	 * @author Martin Mann
	 */
	class SPC_Listing : public StatePairCollector
	{
	public:
		typedef std::list<StatePair> List;
				
	protected:

		//! the internal storage of all added state pairs 
		List statePairList;
	
	public:	
	
		SPC_Listing();
		
		virtual
		~SPC_Listing();
	
		  //! Adds the given State pair to the list.
		  //! @param s the State pair to add
		virtual void add(const StatePair& s);
		
		  //! Adds the given State pair to the list.
		  //! @param s1 the first State of the pair to add 
		  //! @param s2 the second State of the pair to add 
		virtual void add(	const State* const s1
							, const State* const s2 );
		
		  //! Returns number of added State pairs.
		  //! @return the number of added state pairs
		virtual size_t size() const;
		 
		  //! Returns last added State pair.
		  //! @return last added State pair
		virtual const StatePair getLastAdded() const;
		
		///////////  LIST HANDLING  ////////////////////////
	
		  //! Returns a list of all added State pairs.
		  //! @return the StatePair list
		virtual const List& getList() const;
		
	};
	
} // namespace ell


//##############################################################################
//##############################################################################

#include <ostream>
#include <string>

namespace ell {

	/*!
	 * This class counts the number of State pairs added to the collector and
	 * print each to stream. 
	 * 
	 * A pair (s1,s2) is written by 
	 * "s1->toString()+separator+s2->toString()+toAppend"
	 * 
	 * @author Martin Mann
	 */
	class SPC_Outstream : public SPC_Counting 
	{
	protected:
		  //! the stream to write the state pairs to
		std::ostream& out;
		
		  //! additional string written to stream BETWEEN the states of the pair
		const std::string strSep;
		
		  //! additional string written to stream AFTER the state pair
		const std::string strAppend;

	public:
		
		  /*! Construction.
		   * @param out the stream to write to
		   * @param separator the string to use as separator BETWEEN the states
		   *        of the pair
		   * @param toAppend the string to write AFTER the pair to stream as a 
		   *        separator between printed pairs
		   */
		SPC_Outstream(	std::ostream& out
						, const std::string& separator
						, const std::string& toAppend );

		
		virtual
		~SPC_Outstream();
	
		  //! Increments State pair counter and stores a copy this pair as the
		  //! last added.
		  //! @param s the State pair to add
		virtual void add(	const StatePair& s );
		
		  //! Increments State pair counter and stores this pair as the last
		  //! added.
		  //! @param s1 the first State of the pair to add 
		  //! @param s2 the second State of the pair to add 
		virtual void add(	const State* const s1
							, const State* const s2 );
		
	protected:
		
		  //! Writes the last added pair to stream.
		void printToStream();

	};
	
	inline void SPC_Outstream::printToStream() {
		out <<lastAdded_s1->toString()
			<<strSep
			<<lastAdded_s2->toString()
			<<strAppend;
	}

} // namespace ell

//##############################################################################
//##############################################################################


namespace ell
{
	/*!
	 * This class is used to keep track of the number of States pairs
	 * added to the collector and stores the minimal pair added.
	 * 
	 * A base pair (i,j) is considered to be smaller than another (k,l) if 
	 * max(i,j) < max(k,l) according to the State::less operator. 
	 * 
	 * @author Martin Mann
	 */
	class SPC_MinPair : public SPC_Counting
	{
		
	protected:
		
		  //! Whether or not the pairs are added to the collector in ascending
		  //! order such that the first reported pair is the minimal one.
		  //! This reduces the comparison effort enormous!
		const bool pairsAddedInAscOrder;

		  //! the first State of the minimal added State pair 
		State* 	minPair_s1;
		  //! the second State of the minimal added State pair 
		State* 	minPair_s2;
		  //! pointer to the higher of both minPair states according to 
		  //! State::less operator 
		const State*	minPair_max;
		
		using SPC_Counting::lastAdded_s1;
		using SPC_Counting::lastAdded_s2;
		
	public:	
		
		  //! Construction.
		  //! Whether or not the pairs are added to the collector in ascending
		  //! order such that the first reported pair is the minimal one.
		  //! This reduces the comparison effort enormous!
		SPC_MinPair( const bool pairsAddedInAscOrder = false );
		
		virtual
		~SPC_MinPair();
	
		  //! Increments State pair counter and stores a copy this pair as the
		  //! last added.
		  //! @param s the State pair to add
		virtual void add(	const StatePair& s );
		
		  //! Increments State pair counter and stores this pair as the last
		  //! added.
		  //! @param s1 the first State of the pair to add 
		  //! @param s2 the second State of the pair to add 
		virtual void add(	const State* const s1
							, const State* const s2 );
		 
		  //! Returns minimal pair of States added.
		  //! Note: The pointed States will be deleted when the collector is
		  //! destroyed. So a cloning is needed in case the pair will be needed
		  //! later.
		  //! @return minimal added State pair
		virtual const StatePair getMinPair() const;
		
	protected:
		
		  //! delete the minimal added state pair
		void clearMinPair();
		
		  //! checks if the last added state pair is smaller than the current
		  // minimal pair
		void checkMinPair();
		
	};
	
	  // inline defintion 
	inline void SPC_MinPair::clearMinPair() {
		if (minPair_s1 != NULL)
			delete minPair_s1;
		if (minPair_s2 != NULL)
			delete minPair_s2;
	}

	
} // namespace ell


//##############################################################################
//##############################################################################


#endif /*STATEPAIRCOLLECTOR_HH_*/
